# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_log import log as logging

from tacker._i18n import _
from tacker.common import exceptions


LOG = logging.getLogger(__name__)


class GlanceClientException(exceptions.TackerException):
    message = _("%(msg)s")


class ImageCreateWaitFailed(exceptions.TackerException):
    message = _('Create image failed %(reason)s')
