# Copyright 2018 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# openstack infra constants

STACK_CREATE_IN_PROGRESS = "CREATE_IN_PROGRESS"
STACK_CREATE_COMPLETE = "CREATE_COMPLETE"
STACK_UPDATE_IN_PROGRESS = "UPDATE_IN_PROGRESS"
STACK_UPDATE_COMPLETE = "UPDATE_COMPLETE"
STACK_DELETE_IN_PROGRESS = "DELETE_IN_PROGRESS"
STACK_DELETE_COMPLETE = "DELETE_COMPLETE"
