# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""empty message

Revision ID: 4c31092895b8
Revises: 81ffa86020d
Create Date: 2014-08-01 11:48:10.319498

"""

# revision identifiers, used by Alembic.
revision = '4c31092895b8'
down_revision = '81ffa86020d'


def upgrade(active_plugins=None, options=None):
    pass
