# Copyright 2013, 2014 Intel Corporation.
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# key
KEY_ACTION = 'action'
KEY_KWARGS = 'kwargs'
KEY_SCALE = 'scale_stack_id'

# ACTION type
ACTION_CREATE_VNF = 'create_vnf'
ACTION_UPDATE_VNF = 'update_vnf'
ACTION_DELETE_VNF = 'delete_vnf'
ACTION_INSTANTIATE_VNF = 'instantiate_vnf'
ACTION_TERMINATE_VNF = 'terminate_vnf'
ACTION_HEAL_VNF = 'heal_vnf'
ACTION_SCALE_IN_VNF = 'scale_in_vnf'
ACTION_SCALE_OUT_VNF = 'scale_out_vnf'
